﻿/*
 * Weihnachtsstern.c - Scott-Falk Hühn - Version 1.11, 12.12.2021
 *
 * Controller: ATtiny44A, Taktfrequenz 16,0 MHz
 *
 * Belegung der I/O-Ports:
 * A0: Eingang Analog, Poti Farbe Rot
 * A1: Eingang Analog, Poti Farbe Grün
 * A2: Eingang Analog, Poti Farbe Blau
 * A3: Eingang mit Pull-up, Taster
 * A4: Eingang mit Pull-Up, SCK
 * A5: Eingang mit Pull-up, MISO
 * A6: Eingang mit Pull-up, MOSI
 * B2: Ausgang Daten zur LED-Kette
 *
 * Fuse-Bits:
 * SELFPRGEN = [ ]
 * RSTDISBL =  [ ]
 * DWEN =      [ ]
 * SPIEN =     [X]
 * WDTON =     [ ]
 * EESAVE =    [X]
 * BODLEVEL =  2V7
 * CKDIV8 =    [ ]
 * CKOUT =     [ ]
 * SUT_CKSEL = EXTXOSC_8MHZ_XX_16KCK_14CK_65MS
 *
 * EXTENDED = 0xFF
 * HIGH = 0xD5
 * LOW = 0xFF
 */

#define F_CPU 16000000                                  // Taktfrequenz 16,0MHz
#include <avr/io.h>                                     // AVR-I/O-Definitionen laden
#include <avr/interrupt.h>                              // AVR-Interrupt-Routinen laden
#include <avr/eeprom.h>                                 // AVR-EEPROM-Funktionen laden
#include <avr/pgmspace.h>                               // AVR-Flash-ROM-Funktionen laden

#define TIM0_CLK 500                                    // Timer 0 Taktfrequenz = 500Hz (2ms)                                                               |
#define NUM_LEDS 50                                     // Anzahl der LEDs                                                                                  |
#define NUM_EFFS 6                                      // Anzahl der Effekt-LEDs                                                                           | Diese Werte sollten
#define NUM_TOP  4                                      // Nummer der LED an der oberen Spitze                                                              | nicht geändert werden
#define MIN_VAL  40                                     // kleinster möglicher Farbwert beim Farbwechsel-Effekt (Summe von RGB)                             |
#define EFF_MAX  80                                     // maximaler Farbwert für Effekt-LEDs                                                               |

#define INIT_VAL 14                                     // Helligkeitswert  beim Start                                                                      |
#define EFF_SET  20                                     // Helligkeitswert während der Anzeige der Poti-Einstellung                                         |
#define STAR_MIN 10                                     // minimale Anzahl funkelnder Sterne                                                                |
#define STAR_MAX 30                                     // maximale Anzahl funkelnder Sterne                                                                |
const uint8_t CHANGE[]   PROGMEM = { 150,  120,   90};  // Zeitwert für Farbwechsel-Effekt in 3 Geschwindigkeitsstufen (Einheiten von 2ms, maximal 255)     |
const uint8_t EFF_UP[]   PROGMEM = {   6,    5,    4};  // Zeitwert für Effekt-Einblendung in 3 Geschwindigkeitsstufen (Einheiten von 2ms, maximal 255)     | Diese Werte können
const uint8_t EFF_DOWN[] PROGMEM = {   9,    8,    7};  // Zeitwert für Effekt-Ausblendung in 3 Geschwindigkeitsstufen (Einheiten von 2ms, maximal 255)     | geändert werden
const uint16_t DEL_MIN[] PROGMEM = { 300,  250,  200};  // minimale Zeit bis zum nächsten Sternenfunkeln in 3 Geschwindigkeitsstufen (Einheiten von 2ms)    |
const uint16_t DEL_MAX[] PROGMEM = {1400, 1200, 1000};  // maximale Zeit bis zum nächsten Sternenfunkeln in 3 Geschwindigkeitsstufen (Einheiten von 2ms)    |
const uint16_t DEL_EFF[] PROGMEM = {1200, 1000,  850};  // Pausenzeit zwischen den Effekten in 3 Geschwindigkeitsstufen (Einheiten von 2ms)                 |
const uint16_t EFF_TIM[] PROGMEM = { 230,  200,  170};  // Zeitwert für Lauflicht-Effekt in 3 Geschwindigkeitsstufen (Einheiten von 2ms)                    |

static uint8_t led_tab[3 * NUM_LEDS];                   // LED-Farbwert-Tabelle mit 3 Bytes pro LED in der Reihenfolge G-B-R
static uint16_t lfsr;                                   // Zwischenspeicher für Zufallswert
static uint8_t red;                                     // Farbwechsel-Zielwert Rot
static uint8_t green;                                   // Farbwechsel-Zielwert Grün
static uint8_t blue;                                    // Farbwechsel-Zielwert Blau
static uint8_t cur_red;                                 // aktueller Wert Rot
static uint8_t cur_green;                               // aktueller Wert Grün
static uint8_t cur_blue;                                // aktueller Wert Blau
static uint8_t eff_red[NUM_EFFS];                       // Effekt-Wert Rot
static uint8_t eff_green[NUM_EFFS];                     // Effekt-Wert Grün
static uint8_t eff_blue[NUM_EFFS];                      // Effekt-Wert Blau
static uint8_t eff_led1[NUM_EFFS];                      // Effekt-LED-Nummer (0 - NUM_LEDS)
static uint8_t eff_led2[NUM_EFFS];                      // Effekt-LED-Nummer für gespiegeltes Lauflicht (0 - NUM_LEDS)
static uint8_t eff_stat[NUM_EFFS];                      // Effekt-Status (0 = aus/frei, 1 = Einblenden, 2 = Ausblenden)
static uint8_t eff_cnt[NUM_EFFS];                       // Effekt-Warte-Zähler (0-255, entspricht 0-510ms)
volatile uint16_t delay;                                // Verzögerungs-Zähler für Effekte (0-4999, entspricht 0-10s)
volatile uint8_t max_red;                               // Maximalwert Rot (0-63)
volatile uint8_t max_green;                             // Maximalwert Grün (0-63)
volatile uint8_t max_blue;                              // Maximalwert Blau (0-63)
volatile uint8_t int_cyc;                               // Interrupt-Zyklus-Zähler für die Umschaltung der ADC-Eingänge (0-2)
volatile uint8_t int_cnt;                               // Interrupt-Zyklus-Zähler für die LED-Datenausgabe
volatile uint8_t chg_cnt;                               // Interrupt-Zähler für Farbwechsel-Effekt
volatile uint8_t led_flag;                              // LED-Ausgabe-Flag, wird von Interrupt-Service-Routine gesetzt und vom Hauptprogramm ausgewertet
volatile uint8_t but_cnt1;                              // Taster-Zähler 1 (0-24), zählt alle 2ms, nach Erreichen des Endwertes (50ms) wird der Taster gelesen
volatile uint8_t but_cnt2;                              // Taster-Zähler 2 (0-20), zählt bei gedrücktem Taster die Zyklen von but_cnt1 zur Erkennung des langen Tasterdrucks (>1s)
volatile uint8_t but_flag;                              // Taster-Flag: 0 = Taster losgelassen, 1 = Taster gedrückt, 2 = Taster quittiert, 3 = Taster lang gedrückt, 4 = Taster Langdruck quittiert
static uint8_t set_flag;                                // Einstellungs-Flag, wird durch Tasterdruck gesetzt oder gelöscht
static uint8_t eff_num;                                 // Effekt-Nummer (0 = kein Effekt, 1 = Sternenfunkeln, 2 = Lauflicht im Uhrzeigersinn, 3 = Lauflicht gegen den Uhrzeigersinn,
                                                        // 4 = Lauflicht mit Übersprung im Uhrzeigersinn, 5 = Lauflicht mit Übersprung gegen den Uhrzeigersinn, 6 = gespiegeltes Lauflicht)
static uint8_t cur_led1;                                // aktuelle LED-Nummer für Lauflicht-Effekte
static uint8_t cur_led2;                                // aktuelle LED-Nummer für gespiegeltes Lauflicht
static uint8_t led_cnt;                                 // LED-Zähler für funkelnde Sterne und Lauflicht-Effekte
static uint8_t speed;                                   // Geschwindigkeit der Effekte (0-2)

uint8_t e_speed EEMEM;                                  // Geschwindigkeit der Effekte (0-2) im EEPROM

extern void led_out(uint16_t *, uint8_t);               // externe Assembler-Routine zur Datenausgabe an die LEDs (mit Übergabe des Tabellen-Zeigers und der Tabellengröße)

uint16_t rnd(void)                                                    // --- 16-Bit-Zufallszahlengenerator ---
{                                                                     // gefunden auf: https://stackoverflow.com/questions/7602919/how-do-i-generate-random-numbers-without-rand-function
  uint16_t bit;                                                       // Zwischenspeicher
  bit = ((lfsr >> 0) ^ (lfsr >> 2) ^ (lfsr >> 3) ^ (lfsr >> 5)) & 1;  // Zwischenwert ermitteln
  lfsr = ((lfsr >> 1) | (bit << 15)) ^ TCNT1;                         // neuen Zufallswert ermitteln und zusätzlich mit aktuellem Timer-1-Zähler XOR verknüpfen
  return lfsr;                                                        // neuen Wert übergeben
}

uint8_t incdec(uint8_t curr, uint8_t dest)                            // --- Berechnungs-Routine ---
{
  if (curr < dest) curr ++;                                           // wenn aktueller Wert kleiner als Zielwert -> aktuellen Wert um 1 erhöhen
  if (curr > dest) curr --;                                           // wenn aktueller Wert größer als Zielwert -> aktuellen Wert um 1 vermindern
  return curr;                                                        // berechneten aktuellen Wert zurückgeben
}

void eff_led_on(uint8_t eff, uint8_t led1, uint8_t led2)              // --- Eine Effekt-LED einschalten ---
{
  eff_stat[eff] = 1;                                                  // Effekt-LED-Status = 1 (Einblenden) setzen
  eff_cnt[eff] = pgm_read_byte(&EFF_UP[speed]);                       // Effekt-Warte-Zähler neu setzen (Einblend-Wert)
  eff_red[eff] = cur_red;                                             // Effekt-Wert für Rot von aktuellem Farbwechsel-Wert übernehmen
  eff_green[eff] = cur_green;                                         // Effekt-Wert für Grün von aktuellem Farbwechsel-Wert übernehmen
  eff_blue[eff] = cur_blue;                                           // Effekt-Wert für Blau von aktuellem Farbwechsel-Wert übernehmen
  eff_led1[eff] = led1;                                               // aktuelle LED-Nummer für LED 1 setzen
  eff_led2[eff] = led2;                                               // aktuelle LED-Nummer für LED 2 setzen
  if (eff_led1[eff] >= NUM_LEDS) eff_led1[eff] = eff_led1[eff] - NUM_LEDS; // wenn LED-Nummern-Bereich von LED 1 überschritten -> LED-Nummer korrigieren
}

ISR(TIM0_COMPA_vect)                                                  // --- Interrupt-Service-Routine, Timer 0, Output Compare A (500Hz / 2ms) ---
{
  uint8_t i;                                                          // Zähler-Variable und Zwischenspeicher
  i = ADCH / 4;                                                       // ADC-Wert lesen und in Wertebereich 0-63 umrechnen
  if (i == 0) i = 1;                                                  // kleinsten Wert auf 1 setzen
  if (int_cyc == 0) max_red = i;                                      // wenn Interrupt-Zyklus-Zähler = 0 -> ADC-Wert als maximalen Rot-Wert verwenden
  if (max_red == 0) max_red = 1;
  if (int_cyc == 1) max_green = i;                                    // wenn Interrupt-Zyklus-Zähler = 1 -> ADC-Wert als maximalen Grün-Wert verwenden
  if (max_green == 0) max_green = 1;
  if (int_cyc == 2) max_blue = i;                                     // wenn Interrupt-Zyklus-Zähler = 2 -> ADC-Wert als maximalen Blau-Wert verwenden
  if (max_blue == 0) max_blue = 1;
  int_cyc ++;                                                         // Interrupt-Zyklus-Zähler erhöhen
  if (int_cyc > 2) int_cyc = 0;                                       // wenn Endwert überschritten -> zurücksetzen
  ADMUX = int_cyc;                                                    // ADC-Multiplexer auf nächsten ADC-Eingang schalten
  if (chg_cnt > 0) chg_cnt --;                                        // wenn Interrupt-Zähler für Farbwechsel-Effekt > 0 -> vermindern
  if (delay > 0) delay --;                                            // wenn Verzögerung läuft -> vermindern
  for (i = 0; i < NUM_EFFS; i ++)                                     // alle Effekt-LEDs bearbeiten
  {
    if (eff_cnt[i] > 0) eff_cnt[i] --;                                // wenn Effekt-Warte-Zähler läuft -> vermindern
  }
  int_cnt ++;                                                         // Interrupt-Zyklus-Zähler für die LED-Datenausgabe erhöhen
  if (int_cnt > 1) {                                                  // wenn Endwert überschritten
    int_cnt = 0;                                                      // Zähler zurücksetzen
    led_flag = 1;                                                     // LED-Ausgabe-Flag setzen (nur alle 4ms ausgeben)
  }
  but_cnt1 ++;                                                        // Taster-Zähler 1 erhöhen
  if (but_cnt1 > 24)                                                  // wenn Taster-Zähler-1-Endwert überschritten
  {
    but_cnt1 = 0;                                                     // Taster-Zähler 1 zurücksetzen
    if (! (PINA & (1 << PINA3)))                                      // wenn gedrückter Taster erkannt
    {
      if (but_flag == 0) but_flag = 1;                                // wenn Taster-Status "losgelassen" war -> auf "gedrückt" setzen
      if (but_cnt2 < 255) but_cnt2 ++;                                // außerdem wenn Taster-Zähler 2 noch nicht am Endwert -> erhöhen
      if ((but_cnt2 > 20) && (but_flag < 3)) but_flag = 3;            // wenn Taster-Zähler-2-Schwellwert überschritten -> Taster-Status auf "lang gedrückt" setzen
    }
    else                                                              // wenn nicht gedrückter Taster erkannt
    {
      but_flag = 0;                                                   // Taster-Status auf "losgelassen" setzen
      but_cnt2 = 0;                                                   // Taster-Zähler 2 zurücksetzen
    }
  }
}

int main(void)                                                        // ----- Hauptprogramm-Initialisierung -----
{
  uint8_t i, j, k, l;                                                 // allgemeine Variablen für Zähler und Zwischenspeicher

  DDRA = 0;                                                           // PortA nur Eingänge
  PORTA = (1 << PORTA7) | (1 << PORTA6) | (1 << PORTA5) |	(1 << PORTA4) |	(1 << PORTA3); // PortA, Pull-up bei ISP-Anschlüssen setzen
  DDRB = 1 << PORTB2;                                                 // PortB2 auf Ausgang setzen
  PORTB = 0;                                                          // PortB auf Low setzen
  ADCSRA = (1 << ADEN) | (1 << ADSC) | (1 << ADATE) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // ADC, Enable, Start, Auto Trigger, Prescaler 128 (125kHz)
  ADCSRB = 1 << ADLAR;                                                // ADC, 8-Bit-Wert rechtsbündig, Free Running Mode
  DIDR0 = (1 << ADC2D) | (1 << ADC1D) | (1 << ADC0D);                 // ADC, Digital Input Port auf ADC0, ADC1 und ADC2 abschalten
  ADMUX = 0;                                                          // ADC, Multiplexer auf ADC0 (PortA0) setzen
  TCCR0A = 1 << WGM01;                                                // Timer 0, CTC Mode
  TCCR0B = 1 << CS02;                                                 // Timer 0, Vorteiler 256
  OCR0A = F_CPU / 256 / TIM0_CLK - 1;                                 // Timer 0, Output Compare A für 500Hz (2ms) setzen
  TCNT0 = 0;                                                          // Timer 0, Zähler löschen
  TIMSK0 = 1 << OCIE0A;                                               // Timer 0, Output Compare A Interrupt aktivieren
  TCCR1A = 0;                                                         // Timer 1, Normal Mode
  TCCR1B = 1 << CS10;                                                 // Timer 1, kein Vorteiler, Zählung mit 16MHz

  int_cyc = 0;                                                        // Interrupt-Zyklus-Zähler = 0 setzen
  led_flag = 0;                                                       // LED-Ausgabe-Flag löschen
  speed = eeprom_read_byte(&e_speed);                                 // Geschwindigkeit aus dem EEPROM lesen
  if (speed > 2) speed = 1;                                           // wenn ungültiger Wert -> mittlere Geschwindigkeit einstellen
  sei();                                                              // Interrupts aktivieren
  do {} while (int_cyc == 0);                                         // warten solange Interrupt-Zyklus-Zähler = 0 ist  | (einen kompletten
  do {} while (int_cyc != 0);                                         // warten solange Interrupt-Zyklus-Zähler <> 0 ist | Interrupt-Zyklus abwarten)
  led_flag = 0;                                                       // LED-Ausgabe-Flag zunächst löschen
  lfsr = ((max_red & 0x1f) << 10) | ((max_green & 0x1f) << 5) | (max_blue & 0x1f); // Zufallsgenerator mit jeweils 5 Bit der 3 ADC-Werte initialisieren
  red = INIT_VAL;                                                     // Farbwechsel-Zielwert Rot auf Startwert setzen
  green = INIT_VAL;                                                   // Farbwechsel-Zielwert Grün auf Startwert setzen
  blue = INIT_VAL;                                                    // Farbwechsel-Zielwert Blau auf Startwert setzen
  cur_red = INIT_VAL;                                                 // aktuellen Rot-Wert auf Startwert setzen
  cur_green = INIT_VAL;                                               // aktuellen Grün-Wert auf Startwert setzen
  cur_blue = INIT_VAL;                                                // aktuellen Blau-Wert auf Startwert setzen
  eff_num = 7;                                                        // Start-Effekt setzen

  while (1)                                                           // ----- Hauptprogramm-Endlosschleife -----
  {
    if ((cur_red == red) && (cur_green == green) && (cur_blue == blue)) { // wenn Ziel-RGB-Wert erreicht
      do {                                                            // wiederholen
        red = rnd() / (65535 / max_red);                              // neuen Zufallswert für Rot ermitteln
        green = rnd() / (65535 / max_green);                          // neuen Zufallswert für Grün ermitteln
        blue = rnd() / (65535 / max_blue);                            // neuen Zufallswert für Blau ermitteln
        i = 0;                                                        // Merker für das Generieren neuer Zufallszahlen löschen
        if ((red + green + blue) < MIN_VAL) i = 1;                    // wenn neue Farbwerte unter dem Minimalwert -> Merker setzen
        if ((max_red + max_green + max_blue) < (MIN_VAL + 10)) i = 0; // wenn eingestellte Poti-Farbwerte nah am Minimalwert -> Merker wieder löschen
      }
      while (i);                                                      // wiederholen bis sinnvolle Farbwerte vorliegen
    }

    if (chg_cnt == 0) {                                               // wenn Interrupt-Zähler für Farbwechsel abgelaufen
      chg_cnt = pgm_read_byte(&CHANGE[speed]);                        // Interrupt-Zähler neu setzen
      cur_red = incdec(cur_red, red);                                 // neuen Rot-Wert berechnen
      cur_green = incdec(cur_green, green);                           // neuen Grün-Wert berechnen
      cur_blue = incdec(cur_blue, blue);                              // neuen Blau-Wert berechnen
		}

    for (i = 0; i < NUM_EFFS; i ++) {                                 // alle Effekt-LEDs bearbeiten
      if (eff_cnt[i] == 0) {                                          // wenn Effekt-Warte-Zähler abgelaufen
        if (eff_stat[i] == 2) {                                       // wenn Effekt-Status = 2 (Ausblenden)
          if ((eff_red[i] == cur_red) && (eff_green[i] == cur_green) && (eff_blue[i] == cur_blue)) { // wenn aktuelle Farbwerte erreicht
            eff_stat[i] = 0;                                          // Effekt-Status = 0 (aus) setzen
          }
          else {                                                      // wenn aktuelle Farbwerte noch nicht erreicht
            eff_red[i] = incdec(eff_red[i], cur_red);                 // neuen Effekt-Wert für Rot berechnen
            eff_green[i] = incdec(eff_green[i], cur_green);           // neuen Effekt-Wert für Grün berechnen
            eff_blue[i] = incdec(eff_blue[i], cur_blue);              // neuen Effekt-Wert für Blau berechnen
            eff_cnt[i] = pgm_read_byte(&EFF_DOWN[speed]);             // Effekt-Warte-Zähler neu setzen (Ausblend-Wert)
          }
        }
        if (eff_stat[i] == 1) {                                       // wenn Effekt-Status = 1 (Einblenden)
          if ((eff_red[i] == EFF_MAX) && (eff_green[i] == EFF_MAX) && (eff_blue[i] == EFF_MAX)) { // wenn Maximalwert erreicht
            eff_stat[i] = 2;                                          // Effekt-Status = 2 (Ausblenden) setzen
            eff_cnt[i] = pgm_read_byte(&EFF_DOWN[speed]);             // Effekt-Warte-Zähler neu setzen
          }
          else {                                                      // wenn Maximalwert noch nicht erreicht
            eff_red[i] = incdec(eff_red[i], EFF_MAX);                 // neuen Effekt-Wert für Rot berechnen
            eff_green[i] = incdec(eff_green[i], EFF_MAX);             // neuen Effekt-Wert für Grün berechnen
            eff_blue[i] = incdec(eff_blue[i], EFF_MAX);               // neuen Effekt-Wert für Blau berechnen
            eff_cnt[i] = pgm_read_byte(&EFF_UP[speed]);               // Effekt-Warte-Zähler neu setzen (Einblend-Wert)
          }
        }
      }
    }

    if (delay == 0) {                                                 // wenn Verzögerungs-Zähler abgelaufen
      switch (eff_num) {                                              // Verzweigung entsprechend aktueller Effekt-Nummer

        case 0:                                                       // Effekt-Nummer 0 (nur Farbwechsel)
          if (led_cnt == 0) {                                         // wenn LED-Zähler = 0 (Lauflicht abgelaufen oder Systemstart)
            eff_num = 1;                                              // Effekt-Nummer 1 (Sternenfunkeln) setzen
            led_cnt = rnd() / (65535 / (STAR_MAX - STAR_MIN)) + STAR_MIN; // Anzahl der funkelnden Sterne ermitteln und LED-Zähler setzen
            delay = rnd() / (65535 / (pgm_read_word(&DEL_MAX[speed]) - pgm_read_word(&DEL_MIN[speed]))) + pgm_read_word(&DEL_MIN[speed]); // neuen Verzögerungswert ermitteln und Zähler setzen
          }
          else {                                                      // wenn LED-Zähler <> 0 (Sternenfunkeln abgelaufen)
            eff_num = rnd() / (65535 / 5) + 2;                        // Effekt-Nummer zufällig zwischen 2 und 6 ermitteln
            if (eff_num >= 7) eff_num = 6;                            // wenn Bereich überschritten (ungenaue Integer-Arithmetik) -> korrigieren
            k = rnd() / (65535 / 5) * 10;                             // Startposition zufällig ermitteln (eine der 5 Sternspitzen)
            if (k >= 40) k = 40;                                      // wenn Bereich überschritten (ungenaue Integer-Arithmetik) -> korrigieren
            led_cnt = 0;                                              // LED-Zähler zurücksetzen
            cur_led1 = NUM_TOP + k;                                   // Start-LED-Nummer auf erste LED für Vorwärtszählung setzen
            if ((eff_num == 3) || (eff_num == 5)) {                   // wenn Effekt-Nummer 3 oder 5
              cur_led1 = cur_led1 + NUM_LEDS;                         // Start-LED-Nummer auf letzte LED für Rückwärtszählung setzen
            }
            if (eff_num == 6) {                                       // wenn Effekt-Nummer 6
              cur_led2 = NUM_TOP + NUM_LEDS + k;                      // zusätzlich Start-LED-Nummer auf letzte LED für Rückwärtszählung setzen (Spiegel-LED)
            }
            delay = pgm_read_word(&DEL_EFF[speed]);                   // Pause bis zum nächsten Effekt setzen
          }
          break;

        case 1:                                                       // Effekt-Nummer 1 (Sternenfunkeln)
          j = NUM_EFFS;                                               // freie Effekt-LED suchen, Effekt-LED-Nummer zunächst auf ungültigen Wert voreinstellen
          for (i = 0; i < NUM_EFFS; i ++) {                           // alle Effekt-LEDs durchsuchen
            if (j == NUM_EFFS) {                                      // wenn noch keine freie Effekt-LED gefunden
              if (eff_stat[i] == 0) j = i;                            // wenn freie Effekt-LED gefunden -> Effekt-LED-Nummer speichern
            }
          }
          if (j < NUM_EFFS) {                                         // wenn gültige Effekt-LED-Nummer vorhanden (freie Effekt-LED gefunden)
            eff_stat[j] = 1;                                          // Effekt-LED-Status = 1 (Einblenden) setzen
            eff_cnt[j] = pgm_read_byte(&EFF_UP[speed]);               // Effekt-Warte-Zähler neu setzen (Einblend-Wert)
            eff_red[j] = cur_red;                                     // Effekt-Wert für Rot von aktuellem Farbwechsel-Wert übernehmen
            eff_green[j] = cur_green;                                 // Effekt-Wert für Grün von aktuellem Farbwechsel-Wert übernehmen
            eff_blue[j] = cur_blue;                                   // Effekt-Wert für Blau von aktuellem Farbwechsel-Wert übernehmen
            do {
              k = rnd() / (65535 / NUM_LEDS);                         // neue LED-Nummer per Zufall ermitteln
              if (k >= NUM_LEDS) k = NUM_LEDS - 1;                    // wenn Bereich überschritten (ungenaue Integer-Arithmetik) -> korrigieren
              l = 0;                                                  // ermitteln, ob LED-Nummer bereits verwendet wird, Merker = 0 setzen
              for (i = 0; i < NUM_EFFS; i ++) {                       // alle Effekte durchsuchen
                if ((eff_led1[i] == k) && (eff_stat[i] > 0)) l ++;    // wenn LED-Nummer gefunden und Effekt aktiv ist -> Merker erhöhen
              }
            } while (l > 0);                                          // wiederholen bis ermittelte LED-Nummer ok ist (noch nicht verwendet)
            eff_led1[j] = k;                                          // LED-Nummer speichern
            eff_led2[j] = NUM_LEDS;                                   // LED-Nummer für Spiegel-LED löschen, wird hier nicht genutzt
            led_cnt --;                                               // LED-Zähler vermindern
            if (led_cnt != 0) {                                       // wenn noch nicht alle funkelnden Sterne ausgegeben
              delay = rnd() / (65535 / (pgm_read_word(&DEL_MAX[speed]) - pgm_read_word(&DEL_MIN[speed]))) + pgm_read_word(&DEL_MIN[speed]); // neuen Verzögerungswert ermitteln und Zähler setzen
            }
            else {                                                    // wenn alle funkelnden Sterne ausgegeben
              delay = pgm_read_word(&DEL_EFF[speed]);                 // Pause bis zum nächsten Effekt setzen
              eff_num = 0;                                            // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
              led_cnt = 127;                                          // LED-Zähler auf 127 setzen (Markierung für abgelaufenen Effekt 1)
            }
          }
          break;

        case 2:                                                       // Effekt-Nummer 2 (Lauflicht im Uhrzeigersinn)
          j = NUM_EFFS;                                               // freie Effekt-LED suchen, Effekt-LED-Nummer zunächst auf ungültigen Wert voreinstellen
          for (i = 0; i < NUM_EFFS; i ++) {                           // alle Effekt-LEDs durchsuchen
            if (j == NUM_EFFS) {                                      // wenn noch keine freie Effekt-LED gefunden
              if (eff_stat[i] == 0) j = i;                            // wenn freie Effekt-LED gefunden -> Effekt-LED-Nummer speichern
            }
          }
          if (j < NUM_EFFS) {                                         // wenn gültige Effekt-LED-Nummer vorhanden (freie Effekt-LED gefunden)
            eff_led_on(j, cur_led1, NUM_LEDS);                        // Effekt-LED einschalten (Spiegel-LED wird hier nicht verwendet)
            delay = pgm_read_word(&EFF_TIM[speed]);                   // Verzögerung auf Lauflicht-Effekt-Zeit setzen
            cur_led1 ++;                                              // aktuelle LED-Nummer erhöhen
            led_cnt ++;                                               // LED-Zähler erhöhen
            if (led_cnt > NUM_LEDS) {                                 // wenn letzte LED erreicht
              delay = pgm_read_word(&DEL_EFF[speed]);                 // Pause bis zum nächsten Effekt setzen
              eff_num = 0;                                            // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
              led_cnt = 0;                                            // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
            }
          }
          break;

        case 3:                                                       // Effekt-Nummer 3 (Lauflicht gegen den Uhrzeigersinn)
          j = NUM_EFFS;                                               // freie Effekt-LED suchen, Effekt-LED-Nummer zunächst auf ungültigen Wert voreinstellen
          for (i = 0; i < NUM_EFFS; i ++) {                           // alle Effekt-LEDs durchsuchen
            if (j == NUM_EFFS) {                                      // wenn noch keine freie Effekt-LED gefunden
              if (eff_stat[i] == 0) j = i;                            // wenn freie Effekt-LED gefunden -> Effekt-LED-Nummer speichern
            }
          }
          if (j < NUM_EFFS) {                                         // wenn gültige Effekt-LED-Nummer vorhanden (freie Effekt-LED gefunden)
            eff_led_on(j, cur_led1, NUM_LEDS);                        // Effekt-LED einschalten (Spiegel-LED wird hier nicht verwendet)
            delay = pgm_read_word(&EFF_TIM[speed]);                   // Verzögerung auf Lauflicht-Effekt-Zeit setzen
            cur_led1 --;                                              // aktuelle LED-Nummer vermindern
            led_cnt ++;                                               // LED-Zähler erhöhen
            if (led_cnt > NUM_LEDS) {                                 // wenn letzte LED erreicht
              delay = pgm_read_word(&DEL_EFF[speed]);                 // Pause bis zum nächsten Effekt setzen
              eff_num = 0;                                            // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
              led_cnt = 0;                                            // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
            }
          }
          break;

        case 4:                                                       // Effekt-Nummer 4 (Lauflicht mit Übersprung im Uhrzeigersinn)
          j = NUM_EFFS;                                               // freie Effekt-LED suchen, Effekt-LED-Nummer zunächst auf ungültigen Wert voreinstellen
          for (i = 0; i < NUM_EFFS; i ++) {                           // alle Effekt-LEDs durchsuchen
            if (j == NUM_EFFS) {                                      // wenn noch keine freie Effekt-LED gefunden
              if (eff_stat[i] == 0) j = i;                            // wenn freie Effekt-LED gefunden -> Effekt-LED-Nummer speichern
            }
          }
          if (j < NUM_EFFS) {                                         // wenn gültige Effekt-LED-Nummer vorhanden (freie Effekt-LED gefunden)
            eff_led_on(j, cur_led1, NUM_LEDS);                        // Effekt-LED einschalten (Spiegel-LED wird hier nicht verwendet)
            delay = pgm_read_word(&EFF_TIM[speed]);                   // Verzögerung auf Lauflicht-Effekt-Zeit setzen
            cur_led1 ++;                                              // aktuelle LED-Nummer erhöhen
            led_cnt ++;                                               // LED-Zähler erhöhen
            switch (led_cnt) {                                        // LED-Sprünge und Effekt-Ende abhängig vom LED-Zähler steuern
              case 6:  cur_led1 = cur_led1 + 9; break;                // Sprung von LED 5 auf LED 15
              case 17: cur_led1 = cur_led1 + 9; break;                // Sprung von LED 25 auf LED 35
              case 28: cur_led1 = cur_led1 - 41; break;               // Sprung von LED 45 auf LED 5
              case 39: cur_led1 = cur_led1 + 9; break;                // Sprung von LED 15 auf LED 25
              case 50: cur_led1 = cur_led1 + 9; break;                // Sprung von LED 35 auf LED 45
              case 56:
                delay = pgm_read_word(&DEL_EFF[speed]);               // Ende, Pause bis zum nächsten Effekt setzen
                eff_num = 0;                                          // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
                led_cnt = 0;                                          // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
                break;
            }
          }
          break;

        case 5:                                                       // Effekt-Nummer 5 (Lauflicht mit Übersprung gegen den Uhrzeigersinn)
          j = NUM_EFFS;                                               // freie Effekt-LED suchen, Effekt-LED-Nummer zunächst auf ungültigen Wert voreinstellen
          for (i = 0; i < NUM_EFFS; i ++) {                           // alle Effekte-LEDs durchsuchen
            if (j == NUM_EFFS) {                                      // wenn noch keine freie Effekt-LED gefunden
              if (eff_stat[i] == 0) j = i;                            // wenn freie Effekt-LED gefunden -> Effekt-LED-Nummer speichern
            }
          }
          if (j < NUM_EFFS) {                                         // wenn gültige Effekt-LED-Nummer vorhanden (freie Effekt-LED gefunden)
            eff_led_on(j, cur_led1, NUM_LEDS);                        // Effekt-LED einschalten (Spiegel-LED wird hier nicht verwendet)
            delay = pgm_read_word(&EFF_TIM[speed]);                   // Verzögerung auf Lauflicht-Effekt-Zeit setzen
            cur_led1 --;                                              // aktuelle LED-Nummer vermindern
            led_cnt ++;                                               // LED-Zähler erhöhen
            switch (led_cnt) {                                        // LED-Sprünge und Effekte-Ende abhängig vom LED-Zähler steuern
              case 6:  cur_led1 = cur_led1 - 9; break;                // Sprung von LED 45 auf LED 35
              case 17: cur_led1 = cur_led1 - 9; break;                // Sprung von LED 25 auf LED 15
              case 28: cur_led1 = cur_led1 + 41; break;               // Sprung von LED 5 auf LED 45
              case 39: cur_led1 = cur_led1 - 9; break;                // Sprung von LED 35 auf LED 25
              case 50: cur_led1 = cur_led1 - 9; break;                // Sprung von LED 15 auf LED 5
              case 56:
                delay = pgm_read_word(&DEL_EFF[speed]);               // Ende, Pause bis zum nächsten Effekt setzen
                eff_num = 0;                                          // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
                led_cnt = 0;                                          // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
                break;
            }
          }
          break;

        case 6:                                                       // Effekt-Nummer 6 (gespiegeltes Lauflicht)
          j = NUM_EFFS;                                               // freie Effekt-LED suchen, Effekt-LED-Nummer zunächst auf ungültigen Wert voreinstellen
          for (i = 0; i < NUM_EFFS; i ++) {                           // alle Effekt-LEDs durchsuchen
            if (j == NUM_EFFS) {                                      // wenn noch keine freie Effekt-LED gefunden
              if (eff_stat[i] == 0) j = i;                            // wenn freie Effekt-LED gefunden -> Effekt-LED-Nummer speichern
            }
          }
          if (j < NUM_EFFS) {                                         // wenn gültige Effekt-LED-Nummer vorhanden (freie Effekt-LED gefunden)
            eff_led_on(j, cur_led1, cur_led2);                        // Effekt-LED einschalten (zusätzlich Spiegel-LED einschalten)
            if (eff_led2[j] >= NUM_LEDS) eff_led2[j] = eff_led2[j] - NUM_LEDS; // wenn LED-Nummern-Bereich der Spiegel-LED überschritten -> LED-Nummer korrigieren
            delay = pgm_read_word(&EFF_TIM[speed]);                   // Verzögerung auf Lauflicht-Effekt-Zeit setzen
            cur_led1 ++;                                              // aktuelle LED-Nummer erhöhen
            cur_led2 --;                                              // aktuelle LED-Nummer vermindern (Spiegel-LED)
            led_cnt ++;                                               // LED-Zähler erhöhen
            if (led_cnt == 26) {                                      // wenn Treffpunkt erreicht
              delay = pgm_read_word(&DEL_EFF[speed]);                 // Pause bis zur Fortsetzung setzen
              cur_led1 --;                                            // LED-Nummer auf Treffpunkt zurücksetzen
              cur_led2 ++;                                            // LED-Nummer (Spiegel-LED) auf Treffpunkt zurücksetzen
            }
            if (led_cnt > NUM_LEDS + 1) {                             // wenn letzte LED erreicht
              delay = pgm_read_word(&DEL_EFF[speed]);                 // Pause bis zum nächsten Effekt setzen
              eff_num = 0;                                            // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
              led_cnt = 0;                                            // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
            }
          }
          break;

        case 7:                                                       // Effekt-Nummer 7 (Start-Effekt, Aufblitzen der Sternspitzen)
          eff_led_on(0, NUM_TOP, NUM_LEDS);                           // Effekt-LED für obere Spitze einschalten
          eff_led_on(1, NUM_TOP + 10, NUM_LEDS);                      // Effekt-LED für nächste Spitze einschalten
          eff_led_on(2, NUM_TOP + 20, NUM_LEDS);                      // Effekt-LED für nächste Spitze einschalten
          eff_led_on(3, NUM_TOP + 30, NUM_LEDS);                      // Effekt-LED für nächste Spitze einschalten
          eff_led_on(4, NUM_TOP + 40, NUM_LEDS);                      // Effekt-LED für nächste Spitze einschalten
          delay = pgm_read_word(&DEL_EFF[speed]);                     // Ende, Pause bis zum nächsten Effekt setzen
          eff_num = 0;                                                // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
          led_cnt = 0;                                                // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
          break;
      }
    }

    if (set_flag == 0) {                                              // wenn nicht im Einstellungs-Modus -> LED-Farbwert-Tabelle nach laufendem Programm schreiben
      for (i = 0; i < NUM_LEDS; i ++) {                               // zunächst alle LEDs mit dem aktuellen Farbwert setzen
        led_tab[i * 3] = cur_green * cur_green / 64;                  // aktuellen Grün-Wert als quadratischen Wert für alle LEDs setzen
        led_tab[i * 3 + 1] = cur_red * cur_red / 64;                  // aktuellen Rot-Wert als quadratischen Wert für alle LEDs setzen
        led_tab[i * 3 + 2] = cur_blue * cur_blue / 64;                // aktuellen Blau-Wert als quadratischen Wert für alle LEDs setzen
      }
      for (i = 0; i < NUM_EFFS; i ++) {                               // jetzt alle Effekt-LEDs ausgeben
        if (eff_stat[i] > 0) {                                             // wenn die Effekt-LED aktiv ist
          led_tab[eff_led1[i] * 3] = eff_green[i] * eff_green[i] / 64;     // Effekt-Grün-Wert als quadratischen Wert für die entsprechende Effekt-LED setzen
          led_tab[eff_led1[i] * 3 + 1] = eff_red[i] * eff_red[i] / 64;     // Effekt-Rot-Wert als quadratischen Wert für die entsprechende Effekt-LED setzen
          led_tab[eff_led1[i] * 3 + 2] = eff_blue[i] * eff_blue[i] / 64;   // Effekt-Blau-Wert als quadratischen Wert für die entsprechende Effekt-LED setzen
          if (eff_led2[i] < NUM_LEDS) {                                    // wenn zusätzlich die Spiegel-LED aktiv ist
            led_tab[eff_led2[i] * 3] = eff_green[i] * eff_green[i] / 64;   // Effekt-Grün-Wert als quadratischen Wert für die entsprechende LED setzen
            led_tab[eff_led2[i] * 3 + 1] = eff_red[i] * eff_red[i] / 64;   // Effekt-Rot-Wert als quadratischen Wert für die entsprechende LED setzen
            led_tab[eff_led2[i] * 3 + 2] = eff_blue[i] * eff_blue[i] / 64; // Effekt-Blau-Wert als quadratischen Wert für die entsprechende LED setzen
          }
        }
      }
    }
    else {                                                            // wenn im Einstellungs-Modus -> LED-Farbwert-Tabelle nach Poti-Einstellung schreiben
      for (i = 0; i < NUM_LEDS; i ++) {                               // Schleife, LED-Farbwert-Tabelle schreiben
        j = i + NUM_TOP;                                              // LED-Nummer auf obere LED setzen
        if (j >= NUM_LEDS) j = j - NUM_LEDS;                          // wenn Tabellen-Index überschritten -> korrigieren
        if ((max_green * 50 / 63) > i) led_tab[j * 3] = EFF_SET;      // wenn eingestellter Grün-Wert größer als LED-Nummer -> Grün einschalten
        else led_tab[j * 3] = 0;                                      // sonst LED ausschalten
        if ((max_red * 50 / 63) > i) led_tab[j * 3 + 1] = EFF_SET;    // wenn eingestellter Rot-Wert größer als LED-Nummer -> Rot einschalten
        else led_tab[j * 3 + 1] = 0;                                  // sonst LED ausschalten
        if ((max_blue * 50 / 63) > i) led_tab[j * 3 + 2] = EFF_SET;   // wenn eingestellter Blau-Wert größer als LED-Nummer -> Blau einschalten
        else led_tab[j * 3 + 2] = 0;                                  // sonst LED ausschalten
      }
    }

    if (led_flag) {                                                   // wenn LED-Ausgabe-Flag gesetzt ist
      led_out((uint16_t *) led_tab, 3 * NUM_LEDS);                    // Datenausgabe an die LEDs, Übergabe eines Pointers auf die LED-Farb-Tabelle und der Byte-Anzahl
      led_flag = 0;                                                   // LED-Ausgabe-Flag wieder löschen
    }

    if (but_flag == 1) {                                              // wenn der Taster gedrückt ist und noch nicht quittiert wurde
      but_flag = 2;                                                   // Tasterdruck quittieren
      set_flag ++;                                                    // Einstellungs-Flag erhöhen
      if (set_flag > 1) {                                             // wenn Endwert überschritten
        set_flag = 0;                                                 // Einstellungs-Flag zurücksetzen
        cur_red = 0;                                                  // aktuellen Rot-Wert löschen
        cur_green = 0;                                                // aktuellen Grün-Wert löschen
        cur_blue = 0;                                                 // aktuellen Blau-Wert löschen
        for (i = 0; i < NUM_EFFS; i ++) eff_stat[i] = 0;              // Schleife, Status aller Effekte-LEDs auf "aus" setzen
        eff_num = 0;                                                  // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
        led_cnt = 0;                                                  // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
        delay = pgm_read_word(&DEL_EFF[speed]);                       // Pause bis zum nächsten Effekt setzen
      }
    }

    if (but_flag == 3) {                                              // wenn der Taster lang gedrückt wurde
      but_flag = 4;                                                   // langen Tasterdruck quittieren
      set_flag = 0;                                                   // Einstellungs-Flag zurücksetzen
      cur_red = 0;                                                    // aktuellen Rot-Wert löschen
      cur_green = 0;                                                  // aktuellen Grün-Wert löschen
      cur_blue = 0;                                                   // aktuellen Blau-Wert löschen
      speed ++;                                                       // Geschwindigkeitswert erhöhen
      if (speed > 2) speed = 0;                                       // wenn Wertebereich überschritten -> zurücksetzen
      eeprom_write_byte(&e_speed, speed);                             // neuen Geschwindigkeitswert im EEPROM speichern

      for (i = 0; i < (speed + 1); i ++) {                            // Effekt-LEDs entsprechend des Geschwindigkeitswertes anzeigen (1-3)
        eff_stat[i] = 1;                                              // Effekt-LED-Status = 1 (Einblenden) setzen
        eff_cnt[i] = pgm_read_byte(&EFF_UP[speed]);                   // Effekt-Warte-Zähler neu setzen (Einblend-Wert)
        eff_red[i] = 0;                                               // Effekt-Wert für Rot auf 0 setzen
        eff_green[i] = 0;                                             // Effekt-Wert für Grün auf 0 setzen
        eff_blue[i] = 0;                                              // Effekt-Wert für Blau auf 0 setzen
        eff_led1[i] = NUM_TOP + i;                                    // LED-Nummer setzen
        eff_led2[i] = NUM_LEDS;                                       // gespiegelte LED ausschalten
        if (eff_led1[i] >= NUM_LEDS) eff_led1[i] = eff_led1[i] - NUM_LEDS; // wenn LED-Nummern-Bereich überschritten -> LED-Nummer korrigieren
      }
      for (i = (speed + 1); i < NUM_EFFS; i ++) eff_stat[i] = 0;      // Schleife, Status der restlichen Effekt-LEDs auf "aus" setzen
    }

    if (but_flag == 4) {                                              // wenn langer Tasterdruck quittiert wurde (Farbwechsel blockieren, bis Taster losgelassen wird)
      cur_red = 0;                                                    // aktuellen Rot-Wert löschen
      cur_green = 0;                                                  // aktuellen Grün-Wert löschen
      cur_blue = 0;                                                   // aktuellen Blau-Wert löschen
      eff_num = 0;                                                    // Effekt-Nummer auf 0 setzen (Effekt ausschalten)
      led_cnt = 0;                                                    // LED-Zähler auf 0 setzen (Markierung für abgelaufenen Effekt)
      delay = pgm_read_word(&DEL_EFF[speed]);                         // Pause bis zum nächsten Effekt setzen
    }
  }
}
